@@ -15,8 +15,8 @@
 
 #include <stddef.h>
 #include <X11/Xlib.h>
-#include <gtypes.h>
-#include <LTKWidget.h>
+#include "gtypes.h"
+#include "LTKWidget.h"
 
 //------------------------------------------------------------------------
 // LTKScrollbar
