@@ -13,12 +13,15 @@
 #pragma interface
 #endif
 
-#include <gtypes.h>
+#include "gtypes.h"
 
 class GString;
 class GfxState;
 class GfxColorSpace;
+class GfxImageColorMap;
 class Stream;
+class Link;
+class Catalog;
 
 //------------------------------------------------------------------------
 // OutputDev
@@ -33,7 +36,7 @@ public:
   // Destructor.
   virtual ~OutputDev() {}
 
-  //---- get info about output device
+  //----- get info about output device
 
   // Does this device use upside-down coordinates?
   // (Upside-down means (0,0) is the top left corner of the page.)
@@ -63,15 +66,14 @@ public:
   virtual void cvtUserToDev(double ux, double uy, int *dx, int *dy);
 
   //----- link borders
-  virtual void drawLinkBorder(double x1, double y1, double x2, double y2,
-			      double w) {}
+  virtual void drawLink(Link *link, Catalog *catalog) {}
 
   //----- save/restore graphics state
   virtual void saveState(GfxState *state) {}
   virtual void restoreState(GfxState *state) {}
 
   //----- update graphics state
-  virtual void updateAll(GfxState *state) {}
+  virtual void updateAll(GfxState *state);
   virtual void updateCTM(GfxState *state, double m11, double m12,
 			 double m21, double m22, double m31, double m32) {}
   virtual void updateLineDash(GfxState *state) {}
@@ -82,9 +84,11 @@ public:
   virtual void updateLineWidth(GfxState *state) {}
   virtual void updateFillColor(GfxState *state) {}
   virtual void updateStrokeColor(GfxState *state) {}
+  virtual void updateFillOpacity(GfxState *state) {}
+  virtual void updateStrokeOpacity(GfxState *state) {}
 
   //----- update text state
-  virtual void updateFont(GfxState *state) = 0;
+  virtual void updateFont(GfxState *state) {}
   virtual void updateTextMat(GfxState *state) {}
   virtual void updateCharSpace(GfxState *state) {}
   virtual void updateRender(GfxState *state) {}
@@ -95,26 +99,37 @@ public:
   virtual void updateTextShift(GfxState *state, double shift) {}
 
   //----- path painting
-  virtual void stroke(GfxState *state) = 0;
-  virtual void fill(GfxState *state) = 0;
-  virtual void eoFill(GfxState *state) = 0;
+  virtual void stroke(GfxState *state) {}
+  virtual void fill(GfxState *state) {}
+  virtual void eoFill(GfxState *state) {}
 
   //----- path clipping
   virtual void clip(GfxState *state) {}
   virtual void eoClip(GfxState *state) {}
 
   //----- text drawing
+  virtual void beginString(GfxState *state, GString *s) {}
+  virtual void endString(GfxState *state) {}
   virtual void drawChar(GfxState *state, double x, double y,
-			Guchar c) {}
+			double dx, double dy, Guchar c) {}
+  virtual void drawChar16(GfxState *state, double x, double y,
+			  double dx, double dy, int c) {}
   virtual void drawString(GfxState *state, GString *s) {}
+  virtual void drawString16(GfxState *state, GString *s) {}
 
   //----- image drawing
-  virtual void drawImageMask(GfxState *state, Stream *str,
+  virtual void drawImageMask(GfxState *state, Object *ref, Stream *str,
 			     int width, int height, GBool invert,
-			     GBool inlineImg) = 0;
-  virtual void drawImage(GfxState *state, Stream *str, int width,
-			 int height, GfxColorSpace *colorSpace,
-			 GBool inlineImg) = 0;
+			     GBool inlineImg);
+  virtual void drawImage(GfxState *state, Object *ref, Stream *str,
+			 int width, int height, GfxImageColorMap *colorMap,
+			 GBool inlineImg);
+
+#if OPI_SUPPORT
+  //----- OPI functions
+  virtual void opiBegin(GfxState *state, Dict *opiDict);
+  virtual void opiEnd(GfxState *state, Dict *opiDict);
+#endif
 
 private:
 
