@@ -15,9 +15,14 @@
 #include "Array.h"
 #include "Dict.h"
 #include "XRef.h"
+#include "Link.h"
 #include "OutputDev.h"
+#ifndef PDF_PARSER_ONLY
 #include "Gfx.h"
+#include "FormWidget.h"
+#endif
 #include "Error.h"
+
 #include "Params.h"
 #include "Page.h"
 
@@ -27,6 +32,7 @@
 
 PageAttrs::PageAttrs(PageAttrs *attrs, Dict *dict) {
   Object obj1, obj2;
+  double w, h;
 
   // get old/default values
   if (attrs) {
@@ -34,6 +40,10 @@ PageAttrs::PageAttrs(PageAttrs *attrs, Dict *dict) {
     y1 = attrs->y1;
     x2 = attrs->x2;
     y2 = attrs->y2;
+    cropX1 = attrs->cropX1;
+    cropY1 = attrs->cropY1;
+    cropX2 = attrs->cropX2;
+    cropY2 = attrs->cropY2;
     rotate = attrs->rotate;
     attrs->resources.copy(&resources);
   } else {
@@ -43,6 +53,7 @@ PageAttrs::PageAttrs(PageAttrs *attrs, Dict *dict) {
     y1 = 0;
     x2 = 612;
     y2 = 792;
+    cropX1 = cropY1 = cropX2 = cropY2 = 0;
     rotate = 0;
     resources.initNull();
   }
@@ -51,24 +62,57 @@ PageAttrs::PageAttrs(PageAttrs *attrs, Dict *dict) {
   dict->lookup("MediaBox", &obj1);
   if (obj1.isArray() && obj1.arrayGetLength() == 4) {
     obj1.arrayGet(0, &obj2);
-    if (obj2.isInt())
-      x1 = obj2.getInt();
+    if (obj2.isNum())
+      x1 = obj2.getNum();
     obj2.free();
     obj1.arrayGet(1, &obj2);
-    if (obj2.isInt())
-      y1 = obj2.getInt();
+    if (obj2.isNum())
+      y1 = obj2.getNum();
     obj2.free();
     obj1.arrayGet(2, &obj2);
-    if (obj2.isInt())
-      x2 = obj2.getInt();
+    if (obj2.isNum())
+      x2 = obj2.getNum();
     obj2.free();
     obj1.arrayGet(3, &obj2);
-    if (obj2.isInt())
-      y2 = obj2.getInt();
+    if (obj2.isNum())
+      y2 = obj2.getNum();
     obj2.free();
   }
   obj1.free();
 
+  // crop box
+  dict->lookup("CropBox", &obj1);
+  if (obj1.isArray() && obj1.arrayGetLength() == 4) {
+    obj1.arrayGet(0, &obj2);
+    if (obj2.isNum())
+      cropX1 = obj2.getNum();
+    obj2.free();
+    obj1.arrayGet(1, &obj2);
+    if (obj2.isNum())
+      cropY1 = obj2.getNum();
+    obj2.free();
+    obj1.arrayGet(2, &obj2);
+    if (obj2.isNum())
+      cropX2 = obj2.getNum();
+    obj2.free();
+    obj1.arrayGet(3, &obj2);
+    if (obj2.isNum())
+      cropY2 = obj2.getNum();
+    obj2.free();
+  }
+  obj1.free();
+
+  // if the MediaBox is excessively larger than the CropBox,
+  // just use the CropBox
+  limitToCropBox = gFalse;
+  w = 0.25 * (cropX2 - cropX1);
+  h = 0.25 * (cropY2 - cropY1);
+  if (cropX2 > cropX1 &&
+      ((cropX1 - x1) + (x2 - cropX2) > w ||
+       (cropY1 - y1) + (y2 - cropY2) > h)) {
+    limitToCropBox = gTrue;
+  }
+
   // rotate
   dict->lookup("Rotate", &obj1);
   if (obj1.isInt())
@@ -97,7 +141,6 @@ PageAttrs::~PageAttrs() {
 //------------------------------------------------------------------------
 
 Page::Page(int num1, Dict *pageDict, PageAttrs *attrs1) {
-  Dict *resourceDict;
 
   ok = gTrue;
   num = num1;
@@ -105,27 +148,6 @@ Page::Page(int num1, Dict *pageDict, PageAttrs *attrs1) {
   // get attributes
   attrs = attrs1;
 
-  // resources
-  if ((resourceDict = attrs->getResourceDict())) {
-    resourceDict->lookup("Font", &fontDict);
-    if (!(fontDict.isDict() || fontDict.isNull())) {
-      error(-1, "Font resources object (page %d) is wrong type (%s)",
-	    num, fontDict.getTypeName());
-      fontDict.free();
-      goto err4;
-    }
-    resourceDict->lookup("XObject", &xObjDict);
-    if (!(xObjDict.isDict() || xObjDict.isNull())) {
-      error(-1, "XObject resources object (page %d) is wrong type (%s)",
-	    num, xObjDict.getTypeName());
-      xObjDict.free();
-      goto err3;
-    }
-  } else {
-    fontDict.initNull();
-    xObjDict.initNull();
-  }
-
   // annotations
   pageDict->lookupNF("Annots", &annots);
   if (!(annots.isRef() || annots.isArray() || annots.isNull())) {
@@ -147,10 +169,6 @@ Page::Page(int num1, Dict *pageDict, PageAttrs *attrs1) {
 
   return;
 
- err4:
-  fontDict.initNull();
- err3:
-  xObjDict.initNull();
  err2:
   annots.initNull();
  err1:
@@ -160,54 +178,68 @@ Page::Page(int num1, Dict *pageDict, PageAttrs *attrs1) {
 
 Page::~Page() {
   delete attrs;
-  fontDict.free();
-  xObjDict.free();
   annots.free();
   contents.free();
 }
 
-void Page::display(OutputDev *out, int dpi, int rotate) {
+void Page::display(OutputDev *out, double dpi, int rotate,
+		   Links *links, Catalog *catalog) {
+#ifndef PDF_PARSER_ONLY
   Gfx *gfx;
-  Dict *fonts;
-  Dict *xObjects;
-  Object obj1, obj2;
+  Object obj;
+  Link *link;
   int i;
+  FormWidgets *formWidgets;
 
   if (printCommands) {
-    printf("***** MediaBox = ll:%d,%d ur:%d,%d\n",
-	   attrs->getX1(), attrs->getY1(), attrs->getX2(), attrs->getY2());
+    printf("***** MediaBox = ll:%g,%g ur:%g,%g\n",
+	   getX1(), getY1(), getX2(), getY2());
+    if (isCropped()) {
+      printf("***** CropBox = ll:%g,%g ur:%g,%g\n",
+	     getCropX1(), getCropY1(), getCropX2(), getCropY2());
+    }
     printf("***** Rotate = %d\n", attrs->getRotate());
   }
-  if (fontDict.isDict())
-    fonts = fontDict.getDict();
-  else
-    fonts = NULL;
-  if (xObjDict.isDict())
-    xObjects = xObjDict.getDict();
-  else
-    xObjects = NULL;
-  rotate += attrs->getRotate();
-  if (rotate >= 360)
+
+  rotate += getRotate();
+  if (rotate >= 360) {
     rotate -= 360;
-  else if (rotate < 0)
+  } else if (rotate < 0) {
     rotate += 360;
-  gfx = new Gfx(out, num, fonts, xObjects, dpi, attrs->getX1(), attrs->getY1(),
-		attrs->getX2(), attrs->getY2(), rotate);
-  contents.fetch(&obj1);
-  if (obj1.isArray()) {
-    for (i = 0; i < obj1.arrayGetLength(); ++i) {
-      obj1.arrayGet(i, &obj2);
-      if (obj2.isStream())
-	gfx->display(obj2.getStream());
-      else
-	error(-1, "Weird page contents");
-      obj2.free();
-    }
-  } else if (obj1.isStream()) {
-    gfx->display(obj1.getStream());
-  } else {
-    error(-1, "Weird page contents");
   }
-  obj1.free();
+  gfx = new Gfx(out, num, attrs->getResourceDict(),
+		dpi, getX1(), getY1(), getX2(), getY2(), isCropped(),
+		getCropX1(), getCropY1(), getCropX2(), getCropY2(), rotate);
+  contents.fetch(&obj);
+  if (!obj.isNull()) {
+    gfx->display(&obj);
+  }
+  obj.free();
+
+  // draw links
+  if (links) {
+    for (i = 0; i < links->getNumLinks(); ++i) {
+      link = links->getLink(i);
+      out->drawLink(link, catalog);
+    }
+    out->dump();
+  }
+
+  // draw AcroForm widgets
+  //~ need to reset CTM ???
+  formWidgets = new FormWidgets(annots.fetch(&obj));
+  obj.free();
+  if (printCommands && formWidgets->getNumWidgets() > 0) {
+    printf("***** AcroForm widgets\n");
+  }
+  for (i = 0; i < formWidgets->getNumWidgets(); ++i) {
+    formWidgets->getWidget(i)->draw(gfx);
+  }
+  if (formWidgets->getNumWidgets() > 0) {
+    out->dump();
+  }
+  delete formWidgets;
+
   delete gfx;
+#endif
 }
